from framcore import Base, Model
from framcore.aggregators import Aggregator
from framcore.components import Component, Flow, Node


class JulESAggregator(Base):
    """Class for defining and calculating aggregated Model instances based on a clearing Model.

    In JulES, the clearing Model is the main Model we simulate and collect results from. JulES also has short,
    medium and long term price prognosis problems, which needs aggregated Models to be solved efficiently.
    This class helps create and manage these aggregated Models.

    Note:
        - Short term price prognosis Model is aggregated from a Clearing Model and a list of Aggregators.
        - Medium term price prognosis Model is aggregated from the Short term Model and a list of Aggregators.
        - Long term price prognosis Model is aggregated from the medium term Model and a list of Aggregators.
        - Storages must be the same in all aggregations
        - At the moment the short, medium, and long term Models are the same, du to limitations in JulES (TODO)

    """

    def __init__(
        self,
        clearing: Model,
        short: list[Aggregator],
        medium: list[Aggregator],
        long: list[Aggregator],
    ) -> None:
        """Initialize a JulESAggregator instance.

        Args:
            clearing (Model): The clearing Model to aggregate from.
            short (list[Aggregator]): List of aggregations to create the short term price prognosis Model from clearing.
            medium (list[Aggregator]): List of aggregations to create the medium term price prognosis Model from short
            long (list[Aggregator]): List of aggregations to create the long term price prognosis Model from medium.

        """
        self._clearing = clearing
        self._short = short
        self._medium = medium
        self._long = long

        self._short_model: Model | None = None
        self._medium_model: Model | None = None
        self._long_model: Model | None = None

    def get_short_term_model(self) -> Model:
        """Apply defined aggregations for short term Model."""
        if self._short_model is None:
            self._short_model = self._aggregate(self._clearing, self._short)
        return self._short_model

    def get_medium_term_model(self) -> Model:
        """Apply defined aggregations for medium term Model."""
        if self._medium_model is None:
            self._medium_model = self._aggregate(self.get_short_term_model(), self._medium)
        return self._medium_model

    def get_long_term_model(self) -> Model:
        """Apply defined aggregations for long term Model."""
        if self._long_model is None:
            self._long_model = self._aggregate(self.get_medium_term_model(), self._long)
        return self._long_model

    def get_short_term_aggregation_map(self) -> dict[str, set[str] | None]:
        """Get the aggregation map of Components from clearing to short term Model."""
        return self._create_aggregation_map(self._clearing, self._short)

    def get_medium_term_aggregation_map(self) -> dict[str, set[str] | None]:
        """Get the aggregation map of Components from clearing to medium term Model."""
        return self._create_aggregation_map(self._clearing, self._short + self._medium)

    def get_long_term_aggregation_map(self) -> dict[str, set[str] | None]:
        """Get the aggregation map of Components from clearing to long term Model."""
        return self._create_aggregation_map(self._clearing, self._short + self._medium + self._long)

    def get_short_term_graph_map(
        self,
        graph_clearing: dict[str, Component],
        graph_short: dict[str, Component],
    ) -> dict[str, set[str] | None]:
        """Get aggregation map for version of short term Model with graph of Flows and Nodes."""
        return self._get_graph_aggregation_map(
            original_agg_map=self.get_short_term_aggregation_map(),
            clearing=self._clearing,
            graph_clearing=graph_clearing,
            aggregated=self.get_short_term_model(),
            graph_aggregated=graph_short,
        )

    def get_medium_term_graph_map(
        self,
        graph_clearing: dict[str, Component],
        graph_medium: dict[str, Component],
    ) -> dict[str, set[str] | None]:
        """Get aggregation map for version of medium term Model with graph of Flows and Nodes."""
        return self._get_graph_aggregation_map(
            original_agg_map=self.get_medium_term_aggregation_map(),
            clearing=self._clearing,
            graph_clearing=graph_clearing,
            aggregated=self.get_medium_term_model(),
            graph_aggregated=graph_medium,
        )

    def get_long_term_graph_map(
        self,
        graph_clearing: dict[str, Component],
        graph_long: dict[str, Component],
    ) -> dict[str, set[str] | None]:
        """Get aggregation map for version of long term Model with graph of Flows and Nodes."""
        return self._get_graph_aggregation_map(
            original_agg_map=self.get_long_term_aggregation_map(),
            clearing=self._clearing,
            graph_clearing=graph_clearing,
            aggregated=self.get_long_term_model(),
            graph_aggregated=graph_long,
        )

    def assert_equal_storages(
        self,
        simpler_short: dict[str, Component],
        simpler_medium: dict[str, Component],
        simpler_long: dict[str, Component],
    ) -> None:
        """Check that all Nodes with Storages are preserved between short, medium and long term Models.

        Args:
            simpler_short (dict[str, Component]): Short term Model Components.
            simpler_medium (dict[str, Component]): Medium term Model Components.
            simpler_long (dict[str, Component]): Long term Model Components.

        Raises:
            ValueError: If the Models have differing Storages.

        """
        short_storages = self._get_storages(simpler_short)
        medium_storages = self._get_storages(simpler_medium)
        long_storages = self._get_storages(simpler_long)

        if short_storages != medium_storages != long_storages:
            message = "Storages are not equal between short, medium and long term Models."
            unique_short = short_storages - (medium_storages | long_storages)
            unique_medium = medium_storages - (short_storages | long_storages)
            unique_long = long_storages - (short_storages | medium_storages)
            if unique_short:
                message += f"\n - Unique Nodes with Storages in Short Model: {unique_short}"
            if unique_medium:
                message += f"\n - Unique Nodes with Storages in Medium Model: {unique_medium}"
            if unique_long:
                message += f"\n - Unique Nodes with Storages in Long Model: {unique_long}"
            raise ValueError(message)

    def _aggregate(self, model: Model, aggs: list[Aggregator]) -> Model:
        if aggs:
            # works because aggregators should not modify the original components
            # except if disaggregate is called, but we shall only use aggregate
            agg_model = Model()
            agg_model.get_data().update(model.get_data())
        else:
            agg_model = model
        for agg in aggs:
            agg.aggregate(agg_model)
        return agg_model

    def _create_aggregation_map(self, clearing: Model, aggs: list[Aggregator]) -> dict[str, list[str]]:
        """Merge aggregation maps of a list of aggregators from clearing to final."""
        clearing_ids = [name for name, ob in clearing.get_data().items() if isinstance(ob, Component)]
        full_agg_mapping = {k: {v} for k, v in zip(clearing_ids, clearing_ids, strict=True)}

        for agg in aggs:
            agg_map = agg.get_aggregation_map()
            for detailed_id, aggregated_ids in full_agg_mapping.items():
                if not aggregated_ids:  # Component has been deleted by an aggregation.
                    continue

                new_agg_ids = set()
                for agg_id in aggregated_ids:
                    if agg_id not in agg_map:
                        new_agg_ids.add(agg_id)  # left as is
                        continue
                    if not agg_map[agg_id]:
                        # deleted. if all agg_ids are marked deleted, so is the detailed one.
                        continue
                    new_agg_ids |= agg_map[agg_id]

                full_agg_mapping[detailed_id] = new_agg_ids  # empty set signifies deleted component

        return full_agg_mapping

    def _get_graph_aggregation_map(
        self,
        original_agg_map: dict[str, set[str]],
        clearing: Model | dict[str, Component],
        graph_clearing: dict[str, Component],
        aggregated: Model | dict[str, Component],
        graph_aggregated: dict[str, Component],
    ) -> dict[str, set[str]]:
        """Create aggregation map with simpler Component IDs based on an original mapping from clearing to aggregated.

        Use get_top_parent of components to find IDs in original_agg_map then change to the Flow/Node ID.

        Args:
            original_agg_map (dict[str, set[str]]): Mapping between Components of clearing and aggregated Models.
            clearing (dict[str, Component]): Clearing Model with top parents.
            graph_clearing (dict[str, Component]): Clearing Model version with Flows and Nodes. Derived from
                                                     clearing.
            aggregated (dict[str, Component]): Aggregated Model with top parents. Aggregated from clearing.
            graph_aggregated (dict[str, Component]): Aggregated Model version with Flows and Nodes. Derived from
                                                       aggregated.

        Returns:
            dict[str, set[str]]: Mapping between components of simpler clearing and simpler aggregated Models.

        """
        if isinstance(clearing, Model):
            clearing = {k: v for k, v in clearing.get_data().items() if isinstance(v, Component)}
        if isinstance(aggregated, Model):
            aggregated = {k: v for k, v in aggregated.get_data().items() if isinstance(v, Component)}

        self._check_agg_map_compatibility(clearing, aggregated, original_agg_map)

        graph_clearing_map = self._get_top_parent_to_simple(original=clearing, simpler=graph_clearing)
        graph_aggregated_map = self._get_top_parent_to_simple(original=aggregated, simpler=graph_aggregated)
        simple_agg_map = {}

        for clearing_id, agg_ids in original_agg_map.items():
            # the two if statements are there for if we want to map only a subset of the simpler Components.
            if clearing_id in graph_clearing_map:
                if not agg_ids:
                    continue  # choose not to add deleted components. May change this later.
                simple_agg_ids = set()
                for agg_id in agg_ids:
                    if agg_id in graph_aggregated_map:  # Again to allow subset to be mapped
                        simple_agg_ids |= graph_aggregated_map[agg_id]  # add set if simple component ids
                if simple_agg_ids:
                    for graph_clearing_id in graph_clearing_map[clearing_id]:
                        simple_agg_map[graph_clearing_id] = simple_agg_ids

        self._check_agg_map_validity(graph_clearing, graph_aggregated, simple_agg_map)
        return simple_agg_map

    def _check_agg_map_compatibility(
        self,
        clearing: Model | dict[str, Component],
        aggregated: Model | dict[str, Component],
        original_agg_map: dict[str, set[str]],
    ) -> None:
        if set(clearing.keys()) != set(original_agg_map.keys()):
            missing_in_clearing = set(original_agg_map.keys()).difference(clearing.keys())
            extra_in_clearing = set(clearing.keys()).difference(original_agg_map.keys())
            message = (
                "clearing is incompatible with the aggregation mapping between clearing and aggregated Models.\n"
                f"Missing in clearing: {missing_in_clearing}\n"
                f"Extra in clearing: {extra_in_clearing}"
            )
            raise KeyError(message)

        original_agg_map_values = set().union(*(v for v in original_agg_map.values() if v))
        if set(aggregated.keys()) != original_agg_map_values:
            missing_in_aggregated = original_agg_map_values.difference(aggregated.keys())
            extra_in_aggregated = set(aggregated.keys()).difference(original_agg_map_values)
            message = (
                "aggregated is incompatible with the aggregation mapping between clearing and aggregated Models.\n"
                f"Missing in aggregated: {missing_in_aggregated}\n"
                f"Extra in aggregated: {extra_in_aggregated}"
            )
            raise KeyError(message)

    def _check_agg_map_validity(
        self,
        original_components: dict[str, Component],
        aggregated_components: dict[str, Component],
        agg_map: dict[str, set[str] | None],
    ) -> None:
        """Check Flow and Node rules for all mappings in an aggregation map."""
        errors = set()
        for original_id, aggregated_ids in agg_map.items():
            component = original_components[original_id]
            if isinstance(component, Node):
                self._check_node_rules(original_id, component, aggregated_ids, aggregated_components, errors)

            if isinstance(component, Flow) and component.get_startupcost() is not None:
                self._check_flow_rules(original_id, aggregated_ids, aggregated_components, errors)

            if not isinstance(component, (Flow, Node)):
                message = (
                    f"Invalig Model of simpler Components. Must consist of only Flows and Nodes. Found: {component}"
                )
                raise ValueError(message)

        self._report_errors(errors)

    def _check_node_rules(
        self,
        original_id: str,
        node: Node,
        aggregated_ids: set[str] | None,
        aggregated_components: dict[str, Component],
        errors: set[str],
    ) -> None:
        """Check rules for Nodes for a Component ID in an aggregation map.

        A Node on the disaggregated side (keys) must map to exactly one other Node. More keys are alowed to map to the
        same aggregated Node.

        """
        if node.get_storage() is None:
            # Check rules here?
            return

        if aggregated_ids is None:
            e = f"Node with Storage {original_id} was deleted during aggregations. This is not supported in JulES."
            errors.add(e)
            return
        aggregated_storages = set()
        for agg_id in aggregated_ids:
            agg_component = aggregated_components[agg_id]
            if isinstance(agg_component, Node) and agg_component.get_storage() is not None:
                aggregated_storages.add(agg_id)
        if len(aggregated_storages) != 1:
            errors.add(
                f"Node with Storage {original_id} must be connected to exactly one Node with Storage in the "
                f"aggregation map in JulES. Currently connected to: {aggregated_storages}.",
            )

    def _check_flow_rules(
        self,
        original_id: str,
        aggregated_ids: set[str] | None,
        aggregated_components: dict[str, Component],
        errors: set[str],
    ) -> None:
        """Check rules for Flows for a Component ID in an aggregation map.

        A Flow on the disaggregated side (keys) must map to exactly one other Flow. More keys are alowed to map to the
        same aggregated Flow.

        """
        if aggregated_ids is None:
            e = f"Flow with StartUpCost {original_id} was deleted during aggregations. This is not supported in JulES."
            errors.add(e)
            return
        aggregated_flows = set()
        for agg_id in aggregated_ids:
            agg_component = aggregated_components[agg_id]
            if isinstance(agg_component, Flow) and agg_component.get_startupcost() is not None:
                aggregated_flows.add(agg_id)
        if len(aggregated_flows) != 1:
            errors.add(
                f"Flow with StartUpCost {original_id} must be connected to exactly one Flow with StartUpCost in the "
                f"aggregation map in JulES. Currently connected to: {aggregated_flows}.",
            )

    @staticmethod
    def _get_storages(simpler: dict[str, Component]) -> set[str]:
        nodes_with_storages = set()
        for n, c in simpler.items():
            if isinstance(c, Node) and c.get_storage() is not None:
                nodes_with_storages.add(n)
        return nodes_with_storages

    @staticmethod
    def _get_top_parent_to_simple(original: dict[str, Component], simpler: dict[str, Component]) -> dict[str, set[str]]:
        """Map simpler components to their top parent."""
        inv_original = {c: n for n, c in original.items()}
        simpler_map: dict[str, set[str]] = {}

        for simple_id, component in simpler.items():
            top_parent = component.get_top_parent()
            if top_parent is None:
                message = (
                    f"Component {component} with ID {simple_id} has no parents. This means it has not been "
                    "derived from original."
                )
                raise ValueError(message)
            try:
                top_parent_id = inv_original[top_parent]
            except KeyError as e:
                message = (
                    f"Component {top_parent} does not exist in original Model. This means simpler has not been "
                    "derived from original."
                )
                raise KeyError(message) from e
            if top_parent_id in simpler_map:
                # list has been set, wo we add the simple component id to the
                simpler_map[top_parent_id].add(simple_id)
            else:
                simpler_map[top_parent_id] = {simple_id}

        return simpler_map



from datetime import timedelta
from pathlib import Path

import numpy as np
from framcore.attributes import Arrow, Cost, FlowVolume, LevelProfile, Price, StockVolume
from framcore.components import Flow
from framcore.querydbs import QueryDB
from framcore.timeindexes import (
    AverageYearRange,
    FixedFrequencyTimeIndex,
    ProfileTimeIndex,
    SinglePeriodTimeIndex,
)
from numpy.typing import NDArray

from framjules import JulESConfig
from framjules.solve_handler.build_handler.BuildHandler import BuildHandler
from framjules.solve_handler.dataclasses import ComponentInfo, DomainModels, GraphInfos, NodeFlowGraphs
from framjules.solve_handler.JulESNames import JulESNames


class SerialBuildHandler(BuildHandler):
    """Specialized methods for serial simulation."""

    _PROFILE_VALUE_WHEN_ZERO = 1.0

    def __init__(
        self,
        folder: Path,
        config: JulESConfig,
        names: JulESNames,
        domain_models: DomainModels,
        graphs: NodeFlowGraphs,
        graph_infos: GraphInfos,
        db: QueryDB,
    ) -> None:
        """See BuildHandler."""
        super().__init__(folder, config, names, domain_models, graphs, graph_infos, db)

        # some validations spesific to serial simulation
        self._data_period = self.config.get_data_period()
        self._check_type(self._data_period, SinglePeriodTimeIndex)

        start_year, num_years = self.config.get_weather_years()
        self._avg_year_range = AverageYearRange(start_year=start_year, num_years=num_years)
        if self._avg_year_range.is_52_week_years():
            message = "Expected AverageYearRange to have is_52_week_years() == False."
            raise ValueError(message)

        self._is_float32 = self.config.is_float32()

    def get_attribute_level(
        self,
        root_id: str,
        attribute: LevelProfile,
        target_unit: str | None,
    ) -> float:
        """Get level value."""
        is_max_level = False
        value = attribute.get_data_value(
            db=self.db,
            unit=target_unit,
            level_period=self._data_period,
            scenario_horizon=self._avg_year_range,
            is_max_level=is_max_level,
        )
        if value < 0:
            self.send_warning_event(f"Attribute {root_id} of type {type(attribute).__name__} returned {value}")
        return value

    def get_attribute_profile(
        self,
        root_id: str,
        attribute: LevelProfile | Arrow,
        default: float,
        unit: str | None,
        info: ComponentInfo,
    ) -> str | float:
        """Add profile vector to timevectors and return profile_id."""
        if not attribute.has_profile():
            return default

        profile_id = f"{root_id}_profile"

        timeindex = self._find_profile_timeindex(attribute, info.has_storage_resolution)

        vector: NDArray = attribute.get_scenario_vector(
            db=self.db,
            level_period=self._data_period,
            scenario_horizon=timeindex,
            is_float32=self._is_float32,
            unit=unit,
        )
        vector = np.round(vector, decimals=6)  # avoid very small numerical noise, e.g negative prices from jules run

        if np.isnan(vector).any():
            message = (
                f"Profile {profile_id} in time index {timeindex} contains NaN values. "
                "This may indicate a problem with the data or the configuration."
            )
            raise ValueError(message)

        if (vector < 0).any():
            message = (
                f"Profile {profile_id} in time index {timeindex} has negative values. "
                "This may indicate a problem with the data or the configuration."
            )
            raise ValueError(message)

        denominator = vector.mean()

        if denominator == 0:
            vector.fill(self._PROFILE_VALUE_WHEN_ZERO)
        else:
            np.multiply(vector, 1 / denominator, out=vector)

        self.timevectors[timeindex][profile_id] = vector

        return profile_id

    def _find_profile_timeindex(
        self,
        attribute: LevelProfile | Arrow,
        has_storage_resolution: bool,
    ) -> ProfileTimeIndex:
        is_52_week_years = False

        start_year, num_years = self.config.get_weather_years()

        if has_storage_resolution:
            period_duration = timedelta(minutes=self.config.get_time_resolution().get_clearing_storage_minutes())
        else:
            period_duration = timedelta(minutes=self.config.get_time_resolution().get_clearing_market_minutes())

        fallback = ProfileTimeIndex(
            start_year,
            num_years,
            period_duration=self._get_closest_valid_profile_duration(period_duration),
            is_52_week_years=is_52_week_years,
        )

        ix_set: set[FixedFrequencyTimeIndex] = attribute.get_profile_timeindex_set(db=self.db)
        if not all(isinstance(ix, FixedFrequencyTimeIndex) for ix in ix_set):
            return fallback

        is_one_year = all(ix.is_one_year() for ix in ix_set)

        candidate = min((ix for ix in ix_set), key=lambda ix: ix.get_period_duration().total_seconds())

        s_fallback = fallback.get_period_duration().total_seconds()
        s_candidate = candidate.get_period_duration().total_seconds()
        best_duration = (candidate if s_fallback < s_candidate else fallback).get_period_duration()

        if is_one_year:
            return ProfileTimeIndex(
                start_year=candidate.get_start_time().isocalendar().year,
                num_years=1,
                period_duration=self._get_closest_valid_profile_duration(best_duration),
                is_52_week_years=is_52_week_years,
            )
        return ProfileTimeIndex(
            start_year=start_year,
            num_years=num_years,
            period_duration=self._get_closest_valid_profile_duration(best_duration),
            is_52_week_years=is_52_week_years,
        )

    def get_price_level(self, root_id: str, price: Price, info: ComponentInfo) -> float:
        """get_price_level for serial simulation."""
        return self.get_attribute_level(root_id, price, info.unit_price)

    def get_cost_term_level(self, root_id: str, cost_term: Cost, info: ComponentInfo) -> float:
        """get_cost_term_level for serial simulation."""
        return self.get_attribute_level(root_id, cost_term, info.unit_cost)

    def get_capacity_level(self, root_id: str, capacity: StockVolume | FlowVolume, info: ComponentInfo) -> str | float:
        """get_capacity_level for serial simulation. Handles stock or flow based on info."""
        return self.get_attribute_level(root_id, capacity, info.unit_flow if info.is_flow else info.unit_stock)

    def get_coefficient_level(self, root_id: str, arrow: Arrow, info: ComponentInfo) -> str | float:
        """get_coefficient_level for serial simulation."""
        return self.get_attribute_level(root_id, arrow, info.unit_coeffs[arrow.get_node()])

    def get_price_profile(self, root_id: str, price: Price, info: ComponentInfo) -> str | float:
        """get_price_profile for serial simulation."""
        return self.get_attribute_profile(root_id, price, 1.0, info.unit_price, info)

    def get_cost_term_profile(self, root_id: str, cost_term: Cost, info: ComponentInfo) -> str | float:
        """get_cost_term_profile for serial simulation."""
        return self.get_attribute_profile(root_id, cost_term, 1.0, info.unit_cost, info)

    def get_capacity_profile(
        self,
        root_id: str,
        capacity: FlowVolume | StockVolume,
        info: ComponentInfo,
    ) -> str | float:
        """get_capacity_profile for serial simulation."""
        unit = info.unit_flow if info.is_flow else info.unit_stock
        return self.get_attribute_profile(root_id, capacity, 1.0, unit, info)

    def get_rhs_term_level(
        self,
        rhs_term_id: str,
        flow: Flow,
        arrow: Arrow,
        flow_info: ComponentInfo,
    ) -> float:
        """Convert volume (main node) to target node volume.

        This may scale the volume using arrow coefficient,
        e.g. due to transportation loss.

        This may also change unit, if target node belongs to
        different commodity than main node, such as for hydropower.
        """
        coeff_value = self.get_coefficient_level(rhs_term_id, arrow, flow_info)

        volume = flow.get_volume()
        if volume.has_level():
            volume_value = self.get_attribute_level(
                root_id=rhs_term_id,
                attribute=volume,
                target_unit=flow_info.unit_flow,
            )
        else:
            max_cap = flow.get_max_capacity()
            min_cap = flow.get_min_capacity()
            if max_cap is not None and max_cap == min_cap:
                volume_value = self.get_attribute_level(
                    root_id=rhs_term_id,
                    attribute=max_cap,
                    target_unit=flow_info.unit_flow,
                )
            else:
                message = f"{rhs_term_id} is not exogenous"
                raise ValueError(message)

        if coeff_value == 0:
            message = (
                f"Got zero coeff_value for {rhs_term_id}.\n"
                f"volume_value = {volume_value}\n"
                f"coeff_value = {coeff_value}\n"
                f"flow.get_main_node() = {flow.get_main_node()}\n"
                f"arrow.get_node() = {arrow.get_node()}\n"
            )
            raise RuntimeError(message)

        return volume_value / coeff_value  # convert from main_unit to target_unit

    def _rank_profile_timeindex(self, ix: ProfileTimeIndex) -> tuple[bool, float]:
        return ix.is_one_year(), ix.get_period_duration().total_seconds()

    def _select_profile_timeindex(self, *candidates: ProfileTimeIndex) -> ProfileTimeIndex:
        """Select the one with not-is_one_year (if any) and finest period duration."""
        return min(candidates, key=self._rank_profile_timeindex)

    def get_rhs_term_profile(
        self,
        rhs_term_id: str,
        flow: Flow,
        arrow: Arrow,
        flow_info: ComponentInfo,
    ) -> str | float:
        """Create profile (possibly) representing volume_profile * coefficient_profile."""
        volume = flow.get_volume()

        if not volume.has_profile():
            volume = flow.get_max_capacity()

        not_volume_profile = volume is None or not volume.has_profile()
        not_arrow_profile = not arrow.has_profile()
        has_volume_profile = not not_volume_profile
        has_arrow_profile = not not_arrow_profile

        if not_volume_profile and not_arrow_profile:
            return 1.0

        if has_volume_profile and not_arrow_profile:
            return self.get_attribute_profile(
                rhs_term_id,
                attribute=volume,
                default=1.0,
                unit=flow_info.unit_flow,
                info=flow_info,
            )

        if not_volume_profile and has_arrow_profile:
            unit = flow_info.unit_coeffs[arrow.get_node()]
            return self.get_attribute_profile(
                rhs_term_id,
                attribute=arrow,
                default=1.0,
                unit=unit,
                info=flow_info,
            )

        # Here we get both profiles (volume and coefficient) and muliply them
        # together and store the resulting product profile in self.timevectors

        profile_id = f"{rhs_term_id}_profile"

        timeindex = self._select_profile_timeindex(
            self._find_profile_timeindex(volume, flow_info.has_storage_resolution),
            self._find_profile_timeindex(arrow, flow_info.has_storage_resolution),
        )

        x: NDArray = volume.get_scenario_vector(
            db=self.db,
            level_period=self._data_period,
            scenario_horizon=timeindex,
            is_float32=self._is_float32,
            unit=flow_info.unit_flow,
        )

        y: NDArray = arrow.get_scenario_vector(
            db=self.db,
            level_period=self._data_period,
            scenario_horizon=timeindex,
            is_float32=self._is_float32,
            unit=flow_info.unit_coeffs[arrow.get_node()],
        )

        x_mean = x.mean()
        y_mean = y.mean()
        if x_mean == 0 or y_mean == 0:
            x.fill(self._PROFILE_VALUE_WHEN_ZERO)
            return x

        np.multiply(x, 1.0 / x_mean, out=x)
        np.multiply(y, 1.0 / y_mean, out=y)

        np.multiply(x, y, out=x)

        self.timevectors[timeindex][profile_id] = x

        return profile_id



import datetime
import math

from framcore.timeindexes import SinglePeriodTimeIndex

from framjules.solve_handler.config_handler.ConfigHandler import ConfigHandler


class SerialConfigHandler(ConfigHandler):
    """Specialized methods for serial simulation."""

    def adapt_config_to_simulation_mode(self, config: dict) -> None:
        """Add serial simulation mode spesific modifications to config."""
        n = self.names

        data_years, scen_years, num_sim_steps = self.config_get_years_and_num_sim_steps()

        main_config = config[n.MAIN]
        main_config[n.DATA_YEARS] = data_years
        main_config[n.SCENARIO_YEARS] = scen_years
        main_config[n.WEEK_START] = 1
        main_config[n.NUM_SIM_YEARS] = 0
        main_config[n.EXTRA_STEPS] = num_sim_steps

        settings = config[n.PY_JULES_SETTINGS_NAME]
        settings[n.TIME][n.PROB_TIME] = {
            n.NORMAL_TIME: n.FIXED_DATA_TWO_TIME,
            n.PHASE_IN_TIME: n.PHASE_IN_FIXED_DATA_TWO_TIME,
            n.PHASE_IN_DELTA_DAYS: 7,
            n.PHASE_IN_DELTA_STEPS: 7,
        }

    def config_get_years_and_num_sim_steps(self) -> tuple:
        """Get data, scen and sim period config for serial simulation."""
        data_period = self.config.get_data_period()
        assert isinstance(data_period, SinglePeriodTimeIndex)
        data_years: list[int] = [data_period.get_start_time().isocalendar().year]

        first_scen_year, num_scen_years = self.config.get_weather_years()
        first_sim_year, num_sim_years = self.config.get_simulation_years()
        assert first_scen_year <= first_sim_year
        assert first_scen_year + num_scen_years >= first_sim_year + num_sim_years
        scen_years: list[int] = [first_sim_year]

        sim_start = datetime.datetime.fromisocalendar(first_sim_year, 1, 1)
        sim_stop = datetime.datetime.fromisocalendar(first_sim_year + num_sim_years, 1, 1)

        clearing_days = self.config.get_time_resolution().get_clearing_days()
        num_sim_steps = math.ceil((sim_stop - sim_start).total_seconds() / (clearing_days * 24 * 3600))

        return data_years, scen_years, num_sim_steps
